@@ -1,22 +1,25 @@
 from pathlib import Path
 import json
 import logging
+import os.path
 import requests
+
+import httpx
+import jwt
+from requests.auth import AuthBase
+
 import yaml
 from functools import lru_cache
 
-
 import numpy as np
 import pandas as pd
 
-from collections.abc import Mapping
 from pyam.core import IamDataFrame
 from pyam.utils import (
     META_IDX,
     IAMC_IDX,
     isstr,
     pattern_match,
-    DEFAULT_META_INDEX,
     islistable,
 )
 
@@ -24,7 +27,7 @@ logger = logging.getLogger(__name__)
 # set requests-logger to WARNING only
 logging.getLogger("requests").setLevel(logging.WARNING)
 
-_AUTH_URL = "https://db1.ene.iiasa.ac.at/EneAuth/config/v1"
+_AUTH_URL = "https://api.manager.ece.iiasa.ac.at"
 _CITE_MSG = """
 You are connected to the {} scenario explorer hosted by IIASA.
  If you use this data in any published format, please cite the
@@ -36,6 +39,8 @@ You are connected to the {} scenario explorer hosted by IIASA.
 # path to local configuration settings
 DEFAULT_IIASA_CREDS = Path("~").expanduser() / ".local" / "pyam" / "iiasa.yaml"
 
+JWT_DECODE_ARGS = {"verify_signature": False, "verify_exp": True}
+
 
 def set_config(user, password, file=None):
     """Save username and password for the IIASA API connection to a file"""
@@ -45,60 +50,96 @@ def set_config(user, password, file=None):
 
     with open(file, mode="w") as f:
         logger.info(f"Setting IIASA-connection configuration file: {file}")
-        yaml.dump(dict(username=user, password=password), f)
+        yaml.dump(dict(username=user, password=password), f, sort_keys=False)
 
 
-def _get_config(file=None):
+def _read_config(file):
     """Read username and password for IIASA API connection from file"""
-    file = Path(file) if file is not None else DEFAULT_IIASA_CREDS
-    if file.exists():
-        with open(file, "r") as stream:
-            return yaml.safe_load(stream)
+    with open(file, "r") as stream:
+        return yaml.safe_load(stream)
 
 
-def _check_response(r, msg="Trouble with request", error=RuntimeError):
+def _check_response(r, msg="Error connecting to IIASA database", error=RuntimeError):
     if not r.ok:
-        raise error("{}: {}".format(msg, str(r.text)))
+        raise error(f"{msg}: {r.text}")
 
 
-def _get_token(creds, base_url):
-    """Parse credentials and get token from IIASA authentication service"""
+class SceSeAuth(AuthBase):
+    def __init__(self, creds: str = None, auth_url: str = _AUTH_URL):
+        """Connection to the Scenario Services Manager AAC service.
 
-    # try reading default config or parse file
-    if creds is None:
-        creds = _get_config()
-    elif isinstance(creds, Path) or isstr(creds):
-        _creds = _get_config(creds)
-        if _creds is None:
-            logger.error(f"Could not read credentials from `{creds}`")
-        creds = _creds
-    else:
-        msg = (
-            "Passing credentials as clear-text is not allowed. "
-            "Please use `pyam.iiasa.set_config(<user>, <password>)` instead!"
-        )
-        raise DeprecationWarning(msg)
+        Parameters
+        ----------
+        creds : pathlib.Path or str, optional
+            Path to a file with authentication credentials
+        auth_url : str, optionl
+            Url of the authentication service
+        """
+        self.client = httpx.Client(base_url=auth_url, timeout=10.0, http2=True)
+        self.access_token, self.refresh_token = None, None
 
-    # if (still) no creds, get anonymous auth and return
-    if creds is None:
-        url = "/".join([base_url, "anonym"])
-        r = requests.get(url)
-        _check_response(r, "Could not get anonymous token")
-        return r.json(), None
+        if creds is None:
+            if DEFAULT_IIASA_CREDS.exists():
+                self.creds = _read_config(DEFAULT_IIASA_CREDS)
+            else:
+                self.creds = None
+        elif isinstance(creds, Path) or isstr(creds):
+            self.creds = _read_config(creds)
+        else:
+            raise DeprecationWarning(
+                "Passing credentials as clear-text is not allowed. "
+                "Please use `pyam.iiasa.set_config(<user>, <password>)` instead!"
+            )
 
-    # parse creds, write warning
-    if isinstance(creds, Mapping):
-        user, pw = creds["username"], creds["password"]
-    else:
-        user, pw = creds
+        # if no creds, get anonymous token
+        # TODO: explicit token for anonymous login will not be necessary for ixmp-server
+        if self.creds is None:
+            r = self.client.get("/legacy/anonym/")
+            if r.status_code >= 400:
+                raise ValueError("Unknown API error: " + r.text)
+            self.user = None
+            self.access_token = r.json()
 
-    # get user token
-    headers = {"Accept": "application/json", "Content-Type": "application/json"}
-    data = {"username": user, "password": pw}
-    url = "/".join([base_url, "login"])
-    r = requests.post(url, headers=headers, data=json.dumps(data))
-    _check_response(r, "Login failed for user: {}".format(user))
-    return r.json(), user
+        # else get user-token
+        else:
+            self.user = self.creds["username"]
+            self.obtain_jwt()
+
+    def __call__(self):
+        try:
+            # raises jwt.ExpiredSignatureError if token is expired
+            jwt.decode(self.access_token, options=JWT_DECODE_ARGS)
+
+        except jwt.ExpiredSignatureError:
+            self.refresh_jwt()
+
+        return {"Authorization": "Bearer " + self.access_token}
+
+    def obtain_jwt(self):
+        r = self.client.post("/v1/token/obtain/", json=self.creds)
+        if r.status_code == 401:
+            raise ValueError(
+                "Credentials not valid to connect to https://manager.ece.iiasa.ac.at."
+            )
+        elif r.status_code >= 400:
+            raise ValueError("Unknown API error: " + r.text)
+
+        _json = r.json()
+        self.access_token = _json["access"]
+        self.refresh_token = _json["refresh"]
+
+    def refresh_jwt(self):
+        try:
+            # raises jwt.ExpiredSignatureError if token is expired
+            jwt.decode(self.refresh_token, options=JWT_DECODE_ARGS)
+            r = self.client.post(
+                "/v1/token/refresh/", json={"refresh": self.refresh_token}
+            )
+            if r.status_code >= 400:
+                raise ValueError("Unknown API error: " + r.text)
+            self.access_token = r.json()["access"]
+        except jwt.ExpiredSignatureError:
+            self.obtain_jwt()
 
 
 class Connection(object):
@@ -115,7 +156,7 @@ class Connection(object):
         were set using :meth:`pyam.iiasa.set_config`.
         Alternatively, you can provide a path to a yaml file
         with entries of 'username' and 'password'.
-    base_url : str, optional
+    auth_url : str, optional
         custom authentication server URL
 
     Notes
@@ -125,26 +166,29 @@ class Connection(object):
     """
 
     def __init__(self, name=None, creds=None, auth_url=_AUTH_URL):
-        self._auth_url = auth_url
-        self._token, self._user = _get_token(creds, base_url=self._auth_url)
+        self._auth_url = auth_url  # scenario services manager API
+        self._base_url = None  # database connection API
+
+        self.auth = SceSeAuth(creds=creds, auth_url=self._auth_url)
 
         # connect if provided a name
         self._connected = None
         if name:
             self.connect(name)
 
-        if self._user:
-            logger.info(f"You are connected as user `{self._user}`")
+        if self.auth.user is not None:
+            logger.info(f"You are connected as user `{self.auth.user}`")
         else:
             logger.info("You are connected as an anonymous user")
 
     @property
     @lru_cache()
     def _connection_map(self):
-        url = "/".join([self._auth_url, "applications"])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
-        r = requests.get(url, headers=headers)
-        _check_response(r, "Could not get valid connection list")
+        # TODO: application-list will be reimplemented in conjunction with ixmp-server
+        r = self.auth.client.get("legacy/applications", headers=self.auth())
+        if r.status_code >= 400:
+            raise ValueError("Unknown API error: " + r.text)
+
         aliases = set()
         conn_map = {}
         for x in r.json():
@@ -178,26 +222,23 @@ class Connection(object):
             name = self._connection_map[name]
 
         valid = self._connection_map.values()
-        if len(valid) == 0:
-            raise RuntimeError(
-                "No valid connections found for the provided credentials."
+        if name not in valid:
+            raise ValueError(
+                f"You do not have access to instance '{name}' or it does not exist. "
+                "Use `Connection.valid_connections` for a list of accessible services."
             )
 
-        if name not in valid:
-            msg = """
-            {} not recognized as a valid connection name.
-            Choose from one of the supported connections for your user: {}.
-            """
-            raise ValueError(msg.format(name, self._connection_map.keys()))
+        # TODO: config will be reimplemented in conjunction with ixmp-server
+        r = self.auth.client.get(
+            f"legacy/applications/{name}/config", headers=self.auth()
+        )
+        if r.status_code >= 400:
+            raise ValueError("Unknown API error: " + r.text)
 
-        url = "/".join([self._auth_url, "applications", name, "config"])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
-        r = requests.get(url, headers=headers)
-        _check_response(r, "Could not get application information")
         response = r.json()
         idxs = {x["path"]: i for i, x in enumerate(response)}
 
-        self._auth_url = response[idxs["baseUrl"]]["value"]
+        self._base_url = response[idxs["baseUrl"]]["value"]
         # TODO: proper citation (as metadata) instead of link to the about page
         if "uiUrl" in idxs:
             about = "/".join([response[idxs["uiUrl"]]["value"], "#", "about"])
@@ -211,8 +252,43 @@ class Connection(object):
         """Currently connected resource (database API connection)"""
         return self._connected
 
-    def index(self, default=True):
-        """Return the index of models and scenarios in the connected resource
+    @property
+    def meta_columns(self):
+        """Return the list of meta indicators in the connected resource"""
+        url = "/".join([self._base_url, "metadata/types"])
+        r = requests.get(url, headers=self.auth())
+        _check_response(r)
+        return pd.read_json(r.text, orient="records")["name"]
+
+    def _query_index(self, default=True, meta=False, cols=[], **kwargs):
+        # TODO: at present this reads in all data for all scenarios,
+        #  it could be sped up in the future to try to query a subset
+        _default = "true" if default else "false"
+        _meta = "true" if meta else "false"
+        add_url = f"runs?getOnlyDefaultRuns={_default}&includeMetadata={_meta}"
+        url = "/".join([self._base_url, add_url])
+        r = requests.get(url, headers=self.auth())
+        _check_response(r)
+
+        # cast response to dataframe, apply filter by kwargs, and return
+        runs = pd.read_json(r.text, orient="records")
+        if runs.empty:
+            logger.warning("No permission to view model(s) or no scenarios exist.")
+            return pd.DataFrame([], columns=META_IDX + ["version", "run_id"] + cols)
+
+        if kwargs:
+            keep = np.ones(len(runs), dtype=bool)
+            for key, values in kwargs.items():
+                if key not in META_IDX + ["version"]:
+                    raise ValueError(f"Invalid filter: '{key}'")
+                keep_col = pattern_match(pd.Series(runs[key].values), values)
+                keep = np.logical_and(keep, keep_col)
+            return runs[keep]
+        else:
+            return runs
+
+    def index(self, default=True, **kwargs):
+        """Return the index of models and scenarios
 
         Parameters
         ----------
@@ -220,36 +296,13 @@ class Connection(object):
             If `True`, return *only* the default version of a model/scenario.
             Any model/scenario without a default version is omitted.
             If `False`, returns all versions.
+        kwargs
+            Arguments to filer by *model* and *scenario*, `*` can be used as wildcard.
         """
         cols = ["version"] if default else ["version", "is_default"]
-        return self._query_index(default)[META_IDX + cols].set_index(META_IDX)
+        return self._query_index(default, **kwargs)[META_IDX + cols].set_index(META_IDX)
 
-    @lru_cache()
-    def _query_index(self, default=True, meta=False):
-        # TODO: at present this reads in all data for all scenarios,
-        #  it could be sped up in the future to try to query a subset
-        _default = "true" if default else "false"
-        _meta = "true" if meta else "false"
-        add_url = f"runs?getOnlyDefaultRuns={_default}&includeMetadata={_meta}"
-        url = "/".join([self._auth_url, add_url])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
-        r = requests.get(url, headers=headers)
-        _check_response(r)
-
-        # cast response to dataframe and return
-        return pd.read_json(r.text, orient="records")
-
-    @property
-    @lru_cache()
-    def meta_columns(self):
-        """Return the list of meta indicators in the connected resource"""
-        url = "/".join([self._auth_url, "metadata/types"])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
-        r = requests.get(url, headers=headers)
-        _check_response(r)
-        return pd.read_json(r.text, orient="records")["name"]
-
-    def meta(self, default=True, **kwargs):
+    def meta(self, default=True, run_id=False, **kwargs):
         """Return categories and indicators (meta) of scenarios
 
         Parameters
@@ -258,21 +311,25 @@ class Connection(object):
             Return *only* the default version of each scenario.
             Any (`model`, `scenario`) without a default version is omitted.
             If `False`, return all versions.
+        run_id : bool, optional
+            Include "run id" column
+        kwargs
+            Arguments to filer by *model* and *scenario*, `*` can be used as wildcard
         """
-        df = self._query_index(default, meta=True)
+        df = self._query_index(default, meta=True, **kwargs)
+
         cols = ["version"] if default else ["version", "is_default"]
-        if kwargs:
-            if kwargs.get("run_id", False):
-                cols.append("run_id")
+        if run_id:
+            cols.append("run_id")
 
         meta = df[META_IDX + cols]
-        if df.metadata.any():
+        if not meta.empty and df.metadata.any():
             extra_meta = pd.DataFrame.from_records(df.metadata)
             meta = pd.concat([meta, extra_meta], axis=1)
 
-        return meta.set_index(DEFAULT_META_INDEX + ([] if default else ["version"]))
+        return meta.set_index(META_IDX + ([] if default else ["version"]))
 
-    def properties(self, default=True):
+    def properties(self, default=True, **kwargs):
         """Return the audit properties of scenarios
 
         Parameters
@@ -281,17 +338,17 @@ class Connection(object):
             Return *only* the default version of each scenario.
             Any (`model`, `scenario`) without a default version is omitted.
             If :obj:`False`, return all versions.
+        kwargs
+            Arguments to filer by *model* and *scenario*, `*` can be used as wildcard
         """
-        _df = self._query_index(default, meta=True)
         audit_cols = ["cre_user", "cre_date", "upd_user", "upd_date"]
-        audit_mapping = dict([(i, i.replace("_", "ate_")) for i in audit_cols])
         other_cols = ["version"] if default else ["version", "is_default"]
+        cols = audit_cols + other_cols
 
-        return (
-            _df[META_IDX + other_cols + audit_cols]
-            .set_index(META_IDX)
-            .rename(columns=audit_mapping)
-        )
+        _df = self._query_index(default, meta=True, cols=cols, **kwargs)
+        audit_mapping = dict([(i, i.replace("_", "ate_")) for i in audit_cols])
+
+        return _df.set_index(META_IDX).rename(columns=audit_mapping)
 
     def models(self):
         """List all models in the connected resource"""
@@ -304,9 +361,8 @@ class Connection(object):
     @lru_cache()
     def variables(self):
         """List all variables in the connected resource"""
-        url = "/".join([self._auth_url, "ts"])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
-        r = requests.get(url, headers=headers)
+        url = "/".join([self._base_url, "ts"])
+        r = requests.get(url, headers=self.auth())
         _check_response(r)
         df = pd.read_json(r.text, orient="records")
         return pd.Series(df["variable"].unique(), name="variable")
@@ -322,10 +378,9 @@ class Connection(object):
             (possibly leading to duplicate region names for
             regions with more than one synonym)
         """
-        url = "/".join([self._auth_url, "nodes?hierarchy=%2A"])
-        headers = {"Authorization": "Bearer {}".format(self._token)}
+        url = "/".join([self._base_url, "nodes?hierarchy=%2A"])
         params = {"includeSynonyms": include_synonyms}
-        r = requests.get(url, headers=headers, params=params)
+        r = requests.get(url, headers=self.auth(), params=params)
         _check_response(r)
         return self.convert_regions_payload(r.text, include_synonyms)
 
@@ -449,10 +504,8 @@ class Connection(object):
                              variable=['Emissions|CO2', 'Primary Energy'])
 
         """
-        headers = {
-            "Authorization": "Bearer {}".format(self._token),
-            "Content-Type": "application/json",
-        }
+        headers = self.auth().copy()
+        headers["Content-Type"] = "application/json"
 
         # retrieve meta (with run ids) or only index
         if meta:
@@ -463,11 +516,15 @@ class Connection(object):
                 # 'run_id' is required to determine `_args`, dropped later
                 _meta = _meta[set(meta).union(["version", "run_id"])]
         else:
-            _meta = self._query_index(default=default).set_index(DEFAULT_META_INDEX)
+            _meta = self._query_index(default=default).set_index(META_IDX)
+
+        # return nothing if no data exists at all
+        if _meta.empty:
+            return
 
         # retrieve data
         _args = json.dumps(self._query_post(_meta, default=default, **kwargs))
-        url = "/".join([self._auth_url, "runs/bulk/ts"])
+        url = "/".join([self._base_url, "runs/bulk/ts"])
         logger.debug(f"Query timeseries data from {url} with data {_args}")
         r = requests.post(url, headers=headers, data=_args)
         _check_response(r)
@@ -496,13 +553,11 @@ class Connection(object):
 
         # define the index for the IamDataFrame
         if default:
-            index = DEFAULT_META_INDEX
+            index = META_IDX
             data.drop(columns="version", inplace=True)
         else:
-            index = DEFAULT_META_INDEX + ["version"]
-            logger.info(
-                "Initializing an `IamDataFrame` " f"with non-default index {index}"
-            )
+            index = META_IDX + ["version"]
+            logger.info(f"Initializing `IamDataFrame` with non-default index {index}")
 
         # merge meta indicators (if requested) and cast to IamDataFrame
         if meta:
@@ -520,8 +575,8 @@ def read_iiasa(name, default=True, meta=True, creds=None, base_url=_AUTH_URL, **
     Parameters
     ----------
     name : str
-        A valid name of an IIASA scenario explorer instance,
-        see :attr:`pyam.iiasa.Connection.valid_connections`
+        | Name of an IIASA Scenario Explorer database instance.
+        | See :attr:`pyam.iiasa.Connection.valid_connections`.
     default : bool, optional
         Return *only* the default version of each scenario.
         Any (`model`, `scenario`) without a default version is omitted.
@@ -529,7 +584,7 @@ def read_iiasa(name, default=True, meta=True, creds=None, base_url=_AUTH_URL, **
     meta : bool or list of strings, optional
         If `True`, include all meta categories & quantitative indicators
         (or subset if list is given).
-    creds : str, :class:`pathlib.Path`, list-like, or dict, optional
+    creds : str or :class:`pathlib.Path`, optional
         | Credentials (username & password) are not required to access
           any public Scenario Explorer instances (i.e., with Guest login).
         | See :class:`pyam.iiasa.Connection` for details.
@@ -541,3 +596,75 @@ def read_iiasa(name, default=True, meta=True, creds=None, base_url=_AUTH_URL, **
         Arguments for :meth:`pyam.iiasa.Connection.query`
     """
     return Connection(name, creds, base_url).query(default=default, meta=meta, **kwargs)
+
+
+def lazy_read_iiasa(
+    file, name, default=True, meta=True, creds=None, base_url=_AUTH_URL, **kwargs
+):
+    """
+    Try to load data from a local cache, failing that, loads it from the internet.
+
+    Check if the file in a given location is an up-to-date version of an IIASA
+    database. If so, load it. If not, load  data from the IIASA scenario explorer
+    database API and save to that location. Does not check that the previously read
+    version is a complete instance of the database, so if the initial load applies a
+    filter, you will read only data that passes the same filter as well as any
+    additional filter you apply.
+
+    Parameters
+    ----------
+    file : str or :class:`pathlib.Path`
+        The location to test for valid data and save the data if not up-to-date. Must be
+        either xls, xlsx or csv.
+    name : str
+        | Name of an IIASA Scenario Explorer database instance.
+        | See :attr:`pyam.iiasa.Connection.valid_connections`.
+    default : bool, optional
+        Return *only* the default version of each scenario.
+        Any (`model`, `scenario`) without a default version is omitted.
+        If :obj:`False`, return all versions.
+    meta : bool or list of strings, optional
+        If `True`, include all meta categories & quantitative indicators
+        (or subset if list is given).
+    creds : str or :class:`pathlib.Path`, optional
+        | Credentials (username & password) are not required to access
+          any public Scenario Explorer instances (i.e., with Guest login).
+        | See :class:`pyam.iiasa.Connection` for details.
+        | Use :meth:`pyam.iiasa.set_config` to set credentials
+          for accessing private/restricted Scenario Explorer instances.
+    base_url : str
+        Authentication server URL
+    kwargs
+        Arguments for :meth:`pyam.iiasa.Connection.query`
+    """
+
+    file = Path(file)
+    assert file.suffix in [
+        ".csv",
+        ".xlsx",
+        ".xls",
+    ], "We will only read and write to csv, xls and xlsx format."
+    if os.path.exists(file):
+        date_set = pd.to_datetime(os.path.getmtime(file), unit="s")
+        version_info = Connection(name, creds, base_url).properties()
+        latest_new = np.nanmax(pd.to_datetime(version_info["create_date"]))
+        latest_update = np.nanmax(pd.to_datetime(version_info["update_date"]))
+        latest = pd.Series([latest_new, latest_update]).max()
+        if latest < date_set:
+            old_read = IamDataFrame(file)
+            if kwargs:
+                old_read = old_read.filter(**kwargs)
+            logger.info("Database read from file")
+            return old_read
+        else:
+            logger.info("Database out of date and will be re-downloaded")
+    # If we get here, we need to redownload the database
+    new_read = read_iiasa(
+        name, meta=meta, default=default, creds=creds, base_url=base_url, **kwargs
+    )
+    Path(file).parent.mkdir(parents=True, exist_ok=True)
+    if file.suffix == ".csv":
+        new_read.to_csv(file)
+    else:
+        new_read.to_excel(file)
+    return new_read
